import java.awt.image.BufferedImage;

class Engine {
  public Engine() { m_Scene = new Scene(); }

  /**
   * Sets the render target canvas
   */
  public void SetTarget( BufferedImage a_Dest, int a_Width, int a_Height ) {
    // set pixel buffer address & size
    m_Dest = a_Dest;
    m_Width = a_Width;
    m_Height = a_Height;
  }

  public Scene GetScene() {
    return m_Scene;
  }

  int count = 0;
  
  /**
   * Naive ray tracing: Intersects the ray with every primitive
   * in the scene to determine the closest intersection
   */
  public Primitive Raytrace( Ray a_Ray, Color a_Acc, int a_Depth, float a_RIndex, float[] a_Dist ) {
    if (a_Depth > common.TRACEDEPTH) {
      return null;
    }
    // trace primary ray
    a_Dist[0] = 1000000.0f;
    vector3 pi;
    Primitive prim = null;
    int result = 0;
    // find the nearest intersection
    for (Primitive pr: m_Scene.primitives()) {
      int res;
      if ((res = pr.Intersect( a_Ray, a_Dist)) != 0) {
        prim = pr;
        result = res; // 0 = miss, 1 = hit, -1 = hit from inside primitive
      }
    }
    // no hit, terminate ray
    if (prim == null) {
      return null;
    }
    // handle intersection
    if (prim.IsLight()) {
      // we hit a light, stop tracing
      a_Acc.Set( 1, 1, 1 );
    } else {
      // determine color at point of intersection
      pi = a_Ray.origin.add(a_Ray.direction.mul(a_Dist[0]));
      // trace lights
      for (Primitive p: m_Scene.primitives()) {
        if (p.IsLight()) {
          Primitive light = p;
          // handle point light source
          float shade = 1.0f;
          if (light.GetType() == Primitive.Type.SPHERE) {
            vector3 L = ((Sphere)light).GetCentre().sub(pi);
            float[] tdist = new float[1];
            tdist[0] = common.LENGTH( L );
            L.mulAssign(1.0f / tdist[0]);
            Ray r = new Ray( pi.add(L.mul(common.EPSILON)), L );
            for (Primitive pr: m_Scene.primitives()) {
              if ((pr != light) && (pr.Intersect( r, tdist ) != 0)) {
                shade = 0;
                break;
              }
            }
          }
          if (shade > 0) {
             // calculate diffuse shading
            vector3 L = ((Sphere)light).GetCentre().sub(pi);
            common.NORMALIZE( L );
            vector3 N = prim.GetNormal( pi );
            if (prim.GetMaterial().GetDiffuse() > 0) {
//              float dot = DOT(N, L);
              float dot = (N.x*L.x + N.y*L.y + N.z*L.z);
              if (dot > 0) {
                float diff = dot * prim.GetMaterial().GetDiffuse() * shade;
                // add diffuse component to ray color
                a_Acc.addAssign(prim.GetMaterial().GetColor().mul(light.GetMaterial().GetColor()).mul(diff));
              }
            }
            // determine specular component
            if (prim.GetMaterial().GetSpecular() > 0) {
              // point light source: sample once for specular highlight
              vector3 V = a_Ray.direction;
//              vector3 R = L - 2.0f * DOT(L, N) * N;
              vector3 R = L.sub(N.mul(2.0f * (L.x*N.x + L.y*N.y + L.z*N.z)));
//              float dot = DOT(V, R);
              float dot = (V.x*R.x + V.y*R.y + V.z*R.z);
              if (dot > 0) {
                float spec = (float)(Math.pow( dot, 20 ) * prim.GetMaterial().GetSpecular() * shade);
                // add specular component to ray color
                a_Acc.addAssign(light.GetMaterial().GetColor().mul(spec));
              }
            }
          }
        }
      }
      // calculate reflection
      float refl = prim.GetMaterial().GetReflection();
      if (refl > 0.0f && (a_Depth < common.TRACEDEPTH)) {
        vector3 N = prim.GetNormal( pi );
//        vector3 R = a_Ray.direction - 2.0f * DOT( a_Ray.direction, N ) * N;
        vector3 R = a_Ray.direction.sub(N.mul(2.0f * (a_Ray.direction.x*N.x + a_Ray.direction.y*N.y + a_Ray.direction.z*N.z)));
        Color rcol = new Color( 0, 0, 0 );
        float[] dist = new float[1];
        Raytrace( new Ray( pi.add(R.mul(common.EPSILON)), R ), rcol, a_Depth + 1, a_RIndex, dist );
        a_Acc.addAssign(prim.GetMaterial().GetColor().mul(rcol).mul(refl));
      }
      // calculate refraction
      float refr = prim.GetMaterial().GetRefraction();
      if ((refr > 0) && (a_Depth < common.TRACEDEPTH)) {
        float rindex = prim.GetMaterial().GetRefrIndex();
        float n = a_RIndex / rindex;
        vector3 N = prim.GetNormal( pi );
        N.mulAssign((float)result);
//        float cosI = -DOT( N, a_Ray.direction );
        float cosI = -(N.x*a_Ray.direction.x + N.y*a_Ray.direction.y + N.z*a_Ray.direction.z);
        float cosT2 = 1.0f - n * n * (1.0f - cosI * cosI);
        if (cosT2 > 0.0f) {
          vector3 T = a_Ray.direction.mul(n).add(N.mul((float)(n * cosI - Math.sqrt( cosT2 ))));
          Color rcol = new Color( 0, 0, 0 );
          float[] dist = new float[1];
          Raytrace( new Ray( pi.add(T.mul(common.EPSILON)), T ), rcol, a_Depth + 1, rindex, dist );
          // apply Beer's law
          Color absorbance = new Color(prim.GetMaterial().GetColor().mul(0.15f * -dist[0]));
          Color transparency = new Color( (float)Math.exp( absorbance.r() ), (float)Math.exp( absorbance.g() ), (float)Math.exp( absorbance.b() ) );
          a_Acc.addAssign(transparency.mul(rcol));
        }
      }
    }
    // return pointer to primitive hit by primary ray
    return prim;
  }

  /**
   * Initializes the renderer, by resetting the line / tile
   * counters and precalculating some values
   */
  public void InitRender() {
    // set firts line to draw to
    m_CurrLine = 20;
    // screen plane in world space coordinates
    m_WX1 = -4;
    m_WX2 = 4;
    m_WY1 = m_SY = 3;
    m_WY2 = -3;
    // calculate deltas for interpolation
    m_DX = (m_WX2 - m_WX1) / m_Width;
    m_DY = (m_WY2 - m_WY1) / m_Height;
    m_SY += m_CurrLine * m_DY;
    // allocate space to store pointers to primitives for previous line
    m_LastRow = new Primitive[m_Width];
  }

  /**
   * Fires rays in the scene one scanline at a time, from left
   * to right
   */
  public boolean Render() {
    // render scene
    vector3 o = new vector3(0, 0, -5);
    // initialize timer
    long msecs = System.currentTimeMillis();
    // reset last found primitive pointer
    Primitive lastprim = null;

    // render remaining lines
    for ( int y = m_CurrLine; y < (m_Height - 20); y++ ) {
      m_SX = m_WX1;
      // render pixels for current line
      for ( int x = 0; x < m_Width; x++ ) {
        // fire primary rays
        Color acc = new Color(0, 0, 0);
        vector3 dir = new vector3( m_SX, m_SY, 0 ).sub(o);
        common.NORMALIZE( dir );
        Ray r = new Ray( o, dir );
        float[] dist = new float[1];
        Primitive prim = Raytrace( r, acc, 1, 1.0f, dist );
        int red, green, blue;
        if (prim != lastprim) {
          lastprim = prim;
          acc.Set(0, 0, 0);
          for ( int tx = -1; tx < 2; tx++ ) {
            for ( int ty = -1; ty < 2; ty++ ) {
              dir = (new vector3( m_SX + m_DX * tx / 2.0f, m_SY + m_DY * ty / 2.0f, 0 )).sub(o);
              common.NORMALIZE( dir );
              r = new Ray( o, dir );
              dist[0] = 0;
              prim = Raytrace( r, acc, 1, 1.0f, dist );
            }
          }
          red = (int)(acc.r() * (256.0 / 9.0));
          green = (int)(acc.g() * (256.0 / 9.0));
          blue = (int)(acc.b() * (256.0 / 9.0));
        } else {
          red = (int)(acc.r() * 256);
          green = (int)(acc.g() * 256);
          blue = (int)(acc.b() * 256);
        }
        if (red > 255) {
          red = 255;
        }
        if (green > 255) {
          green = 255;
        }
        if (blue > 255) {
          blue = 255;
        }
        m_Dest.setRGB(x, y, (red << 16) + (green << 8) + blue);
        m_SX += m_DX;
      }
      m_SY += m_DY;
      // see if we've been working to long already
      if ((System.currentTimeMillis() - msecs) > 100) {
        // return control to windows so the screen gets updated
        m_CurrLine = y + 1;
        return false;
      }
    }
    // all done
    return true;
  }


  // renderer data
  protected float m_WX1, m_WY1, m_WX2, m_WY2, m_DX, m_DY, m_SX, m_SY;
  protected Scene m_Scene;
  protected BufferedImage m_Dest;
  protected int m_Width, m_Height, m_CurrLine;
  protected Primitive[] m_LastRow;
}



